##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import json
import logging
from collections.abc import Sequence
from typing import Any

import pytket.circuit as tk
from phir.model import PHIRModel
from pytket.circuit.logic_exp import RegWiseOp
from pytket.unit_id import UnitID

from .sharding.shard import Cost, Ordering, ShardLayer

logger = logging.getLogger(__name__)

UINTMAX = 2**32 - 1

tket_gate_to_phir = {
    tk.OpType.Reset:    "Init",
    tk.OpType.Measure:  "Measure",
    tk.OpType.noop:     "I",

    tk.OpType.CX:       "CX",
    tk.OpType.CY:       "CY",
    tk.OpType.CZ:       "CZ",
    tk.OpType.H:        "H",
    tk.OpType.PhasedX:  "R1XY",
    tk.OpType.Rx:       "RX",
    tk.OpType.Ry:       "RY",
    tk.OpType.Rz:       "RZ",
    tk.OpType.S:        "SZ",
    tk.OpType.Sdg:      "SZdg",
    tk.OpType.SWAP:     "SWAP",
    tk.OpType.SX:       "SX",
    tk.OpType.SXdg:     "SXdg",
    tk.OpType.T:        "T",
    tk.OpType.Tdg:      "Tdg",
    tk.OpType.TK2:      "R2XXYYZZ",
    tk.OpType.U1:       "RZ",
    tk.OpType.V:        "SX",
    tk.OpType.Vdg:      "SXdg",
    tk.OpType.X:        "X",
    tk.OpType.XXPhase:  "RXX",
    tk.OpType.Y:        "Y",
    tk.OpType.YYPhase:  "RYY",
    tk.OpType.Z:        "Z",
    tk.OpType.ZZMax:    "SZZ",
    tk.OpType.ZZPhase:  "RZZ",
}  # fmt: skip


def arg_to_bit(arg: UnitID) -> list[str | int]:
    """Convert tket arg to Bit."""
    return [arg.reg_name, arg.index[0]]


def assign_cop(into: str | list[str | int], what: Sequence[int]) -> dict[str, Any]:
    """PHIR for assign classical operation."""
    return {
        "cop": "=",
        "returns": [into],
        "args": what,
    }


def convert_subcmd(op: tk.Op, cmd: tk.Command) -> dict[str, Any]:
    """Return PHIR dict give op and its arguments."""
    if op.is_gate():
        try:
            gate = tket_gate_to_phir[op.type]
        except KeyError:
            logging.exception("Gate %s unsupported by PHIR", op.get_name())
            raise
        angles = (op.params, "pi") if op.params else None
        qop: dict[str, Any]
        match gate:
            case "Measure":
                qop = {
                    "qop": gate,
                    "returns": [arg_to_bit(cmd.bits[0])],
                    "args": [arg_to_bit(cmd.args[0])],
                }

            case "R2XXYYZZ":  # three-qubit gate
                qop = {
                    "qop": gate,
                    "angles": angles,
                    "args": [
                        [
                            arg_to_bit(cmd.qubits[0]),
                            arg_to_bit(cmd.qubits[1]),
                            arg_to_bit(cmd.qubits[2]),
                        ]
                    ],
                }

            case (
                "CX"
                | "CY"
                | "CZ"
                | "RXX"
                | "RYY"
                | "RZZ"
                | "SXX"
                | "SXXdg"
                | "SYY"
                | "SYYdg"
                | "SZZ"
                | "SZZdg"
                | "SWAP"
            ):  # two-qubit gates
                qop = {
                    "qop": gate,
                    "angles": angles,
                    "args": [[arg_to_bit(cmd.qubits[0]), arg_to_bit(cmd.qubits[1])]],
                }
            case _:  # single-qubit gates
                qop = {
                    "qop": gate,
                    "angles": angles,
                    "args": [arg_to_bit(cmd.qubits[0])],
                }
        return qop

    match op:  # non-quantum op
        case tk.SetBitsOp():
            return assign_cop(arg_to_bit(cmd.bits[0]), op.values)

        case _:
            # TODO(kartik): NYI
            # https://github.com/CQCL/pytket-phir/issues/25
            raise NotImplementedError


def append_cmd(cmd: tk.Command, ops: list[dict[str, Any]]) -> None:
    """Convert a pytket command to a PHIR command and append to `ops`.

    Args:
        cmd: pytket command obtained from pytket-phir
        ops: the list of ops to append to
    """
    ops.append({"//": str(cmd)})
    if cmd.op.is_gate():
        ops.append(convert_subcmd(cmd.op, cmd))
    else:
        op: dict[str, Any] | None = None
        match cmd.op:
            case tk.SetBitsOp():
                op = convert_subcmd(cmd.op, cmd)

            case tk.BarrierOp():
                # TODO(kartik): confirm with Ciaran/spec
                # https://github.com/CQCL/phir/blob/main/spec.md
                logger.debug("Skipping Barrier instruction")

            case tk.Conditional():  # where the condition is equality check
                op = {
                    "block": "if",
                    "condition": {
                        "cop": "==",
                        "args": [
                            arg_to_bit(cmd.args[0])
                            if cmd.op.width == 1
                            else cmd.args[0].reg_name,
                            cmd.op.value,
                        ],
                    },
                    "true_branch": [convert_subcmd(cmd.op.op, cmd)],
                }

            case tk.RangePredicateOp():  # where the condition is a range
                cond: dict[str, Any]
                match cmd.op.lower, cmd.op.upper:
                    case l, u if l == u:
                        cond = {
                            "cop": "==",
                            "args": [cmd.args[0].reg_name, u],
                        }
                    case l, u if u == UINTMAX:
                        cond = {
                            "cop": ">=",
                            "args": [cmd.args[0].reg_name, l],
                        }
                    case 0, u:
                        cond = {
                            "cop": "<=",
                            "args": [cmd.args[0].reg_name, u],
                        }
                op = {
                    "block": "if",
                    "condition": cond,
                    "true_branch": [assign_cop(arg_to_bit(cmd.bits[0]), [1])],
                }
            case tk.ClassicalExpBox():
                exp = cmd.op.get_exp()
                match exp.op:
                    case RegWiseOp.XOR:
                        cop = "^"
                    case RegWiseOp.ADD:
                        cop = "+"
                    case RegWiseOp.SUB:
                        cop = "-"
                    case RegWiseOp.MUL:
                        cop = "*"
                    case RegWiseOp.DIV:
                        cop = "/"
                    case RegWiseOp.LSH:
                        cop = "<<"
                    case RegWiseOp.RSH:
                        cop = ">>"
                    case RegWiseOp.EQ:
                        cop = "=="
                    case RegWiseOp.NEQ:
                        cop = "!="
                    case RegWiseOp.LT:
                        cop = "<"
                    case RegWiseOp.GT:
                        cop = ">"
                    case RegWiseOp.LEQ:
                        cop = "<="
                    case RegWiseOp.GEQ:
                        cop = ">="
                    case RegWiseOp.NOT:
                        cop = "~"
                    case other:
                        logging.exception("Unsupported classical operator %s", other)
                        raise ValueError
                op = {
                    "cop": cop,
                    "args": [arg["name"] for arg in exp.to_dict()["args"]],
                }
            case m:
                raise NotImplementedError(m)
        if op:
            ops.append(op)


def genphir(
    inp: list[tuple[Ordering, ShardLayer, Cost]], *, machine_ops: bool = True
) -> str:
    """Convert a list of shards to the equivalent PHIR.

    Args:
        inp: list of shards
        machine_ops: whether to include machine ops
    """
    phir: dict[str, Any] = {
        "format": "PHIR/JSON",
        "version": "0.1.0",
        "metadata": {"source": "pytket-phir"},
    }
    ops: list[dict[str, Any]] = []

    qbits = set()
    cbits = set()
    for _orders, shard_layer, layer_cost in inp:
        for shard in shard_layer:
            qbits |= shard.qubits_used
            cbits |= shard.bits_read | shard.bits_written
            for sub_commands in shard.sub_commands.values():
                for sc in sub_commands:
                    append_cmd(sc, ops)
            append_cmd(shard.primary_command, ops)
        if machine_ops:
            ops.append(
                {
                    "mop": "Transport",
                    "duration": (layer_cost, "ms"),
                },
            )

    # TODO(kartik): this may not always be accurate
    # https://github.com/CQCL/pytket-phir/issues/24
    qvar_dim: dict[str, int] = {}
    for qbit in qbits:
        qvar_dim.setdefault(qbit.reg_name, 0)
        qvar_dim[qbit.reg_name] += 1

    cvar_dim: dict[str, int] = {}
    for cbit in cbits:
        cvar_dim.setdefault(cbit.reg_name, 0)
        cvar_dim[cbit.reg_name] += 1

    decls: list[dict[str, str | int]] = [
        {
            "data": "qvar_define",
            "data_type": "qubits",
            "variable": qvar,
            "size": dim,
        }
        for qvar, dim in qvar_dim.items()
    ]

    decls += [
        {
            "data": "cvar_define",
            "data_type": "u32",
            "variable": cvar,
            "size": dim,
        }
        for cvar, dim in cvar_dim.items()
    ]

    phir["ops"] = decls + ops
    PHIRModel.model_validate(phir)
    return json.dumps(phir)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import logging
from typing import cast

from pytket.circuit import Circuit, Command, Conditional, Op, OpType
from pytket.unit_id import Bit, UnitID

from .shard import Shard

NOT_IMPLEMENTED_OP_TYPES = [OpType.CircBox, OpType.WASM]

SHARD_TRIGGER_OP_TYPES = [
    OpType.Measure,
    OpType.Reset,
    OpType.Barrier,
    OpType.SetBits,
    OpType.ClassicalExpBox,  # some classical operations are rolled up into a box
    OpType.RangePredicate,
    OpType.ExplicitPredicate,
]

logger = logging.getLogger(__name__)


def _is_command_global_phase(command: Command) -> bool:
    return command.op.type == OpType.Phase or (
        command.op.type == OpType.Conditional
        and cast(Conditional, command.op).op.type == OpType.Phase
    )


class Sharder:
    """The Sharder class.

    Responsible for taking in a circuit in TKET representation
    and converting it into shards that can be subsequently handled in the
    compilation pipeline.
    """

    def __init__(self, circuit: Circuit) -> None:
        """Create Sharder object.

        Args:
            circuit: tket Circuit
        """
        self._circuit = circuit
        self._pending_commands: dict[UnitID, list[Command]] = {}
        self._shards: list[Shard] = []
        logger.debug("Sharder created for circuit %s", self._circuit)

    def shard(self) -> list[Shard]:
        """Performs sharding algorithm on the circuit the Sharder was initialized with.

        Returns:
            list of Shards needed to schedule
        """
        logger.debug("Sharding beginning")
        commands = self._circuit.get_commands()

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("All commands:")
            for command in commands:
                logger.debug(command)

        for command in commands:
            self._process_command(command)
        self._cleanup_remaining_commands()

        logger.debug("Shard output:")
        for shard in self._shards:
            logger.debug(shard)
        return self._shards

    def _process_command(self, command: Command) -> None:
        """Handles a command per the type and the extant context within the Sharder.

        Args:
            command: tket command (operation, bits, etc)
        """
        logger.debug(
            "Processing command: %s of type %s with args: %s",
            command.op,
            command.op.type,
            command.args,
        )
        if command.op.type in NOT_IMPLEMENTED_OP_TYPES:
            msg = f"OpType {command.op.type} not supported!"
            raise NotImplementedError(msg)

        if _is_command_global_phase(command):
            logger.debug("Ignoring global Phase gate")
            return

        if self.should_op_create_shard(command.op):
            logger.debug("Building shard for command: %s", command)
            self._build_shard(command)
        else:
            self._add_pending_sub_command(command)

    def _build_shard(self, command: Command) -> None:
        """Builds a shard.

        Creates a Shard object given the extant sharding context and the schedulable
        Command object passed in, and appends it to the Shard list.

        Args:
            command: tket command (operation, bits, etc)
        """
        # Rollup any sub commands (SQ gates) that interact with the same qubits
        sub_commands: dict[UnitID, list[Command]] = {}
        for key in (
            key for key in list(self._pending_commands) if key in command.qubits
        ):
            sub_commands[key] = self._pending_commands.pop(key)

        all_commands = [command]
        for sub_command_list in sub_commands.values():
            all_commands.extend(sub_command_list)

        qubits_used = set(command.qubits)
        bits_written = set(command.bits)
        bits_read: set[Bit] = set()

        for sub_command in all_commands:
            bits_written.update(sub_command.bits)
            bits_read.update(
                set(filter(lambda x: isinstance(x, Bit), sub_command.args)),  # type: ignore [misc, arg-type]
            )

        # Handle dependency calculations
        depends_upon: set[int] = set()
        for shard in self._shards:
            # Check qubit dependencies (R/W implicitly) since all commands
            # on a given qubit need to be ordered as the circuit dictated
            if not shard.qubits_used.isdisjoint(command.qubits):
                logger.debug("...adding shard dep %s -> qubit overlap", shard.ID)
                depends_upon.add(shard.ID)
            # Check classical dependencies, which depend on writing and reading
            # hazards: RAW, WAW, WAR
            # NOTE: bits_read will include bits_written in the current impl

            # Check for write-after-write (changing order would change final value)
            # by looking at overlap of bits_written
            elif not shard.bits_written.isdisjoint(bits_written):
                logger.debug("...adding shard dep %s -> WAW", shard.ID)
                depends_upon.add(shard.ID)

            # Check for read-after-write (value seen would change if reordered)
            elif not shard.bits_written.isdisjoint(bits_read):
                logger.debug("...adding shard dep %s -> RAW", shard.ID)
                depends_upon.add(shard.ID)

            # Check for write-after-read (no reordering or read is changed)
            elif not shard.bits_written.isdisjoint(bits_read):
                logger.debug("...adding shard dep %s -> WAR", shard.ID)
                depends_upon.add(shard.ID)

        shard = Shard(
            command,
            sub_commands,
            qubits_used,
            bits_written,
            bits_read,
            depends_upon,
        )
        self._shards.append(shard)
        logger.debug("Appended shard: %s", shard)

    def _cleanup_remaining_commands(self) -> None:
        remaining_qubits = [k for k, v in self._pending_commands.items() if v]
        for qubit in remaining_qubits:
            self._circuit.add_barrier([qubit])
            # Easiest way to get to a command, since there's no constructor. Could
            # create an entire orphan circuit with the matching qubits and the barrier
            # instead if this has unintended consequences
            barrier_command = self._circuit.get_commands()[-1]
            self._build_shard(barrier_command)

    def _add_pending_sub_command(self, command: Command) -> None:
        """Adds a pending command.

        Adds a pending sub command to the buffer to be flushed when a schedulable
        operation creates a Shard.

        Args:
            command:  tket command (operation, bits, etc)
        """
        qubit_key = command.qubits[0]
        if qubit_key not in self._pending_commands:
            self._pending_commands[qubit_key] = []
        self._pending_commands[qubit_key].append(command)
        logger.debug("Adding pending command %s", command)

    @staticmethod
    def should_op_create_shard(op: Op) -> bool:
        """Decide whether to create a shard.

        This includes non-gate operations like measure/reset as well as 2-qubit gates.

        Args:
            op: operation

        Returns:
            `True` if the operation is one that should result in shard creation
        """
        return (
            op.type in (SHARD_TRIGGER_OP_TYPES)
            or (
                op.type == OpType.Conditional
                and cast(Conditional, op).op.type in (SHARD_TRIGGER_OP_TYPES)
            )
            or (op.is_gate() and op.n_qubits > 1)
        )



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import logging
from typing import TYPE_CHECKING

from rich import print

from phir.model import PHIRModel
from pytket.circuit import Circuit

from .phirgen import genphir
from .place_and_route import place_and_route
from .qtm_machine import QTM_MACHINES_MAP, QtmMachine
from .rebasing.rebaser import rebase_to_qtm_machine
from .sharding.sharder import Sharder

if TYPE_CHECKING:
    from .machine import Machine

logger = logging.getLogger(__name__)


def pytket_to_phir(
    circuit: Circuit,
    qtm_machine: QtmMachine | None = None,
) -> str:
    """Converts a pytket circuit into its PHIR representation.

    This can optionally include rebasing against a Quantinuum machine architecture.

    :param circuit: Circuit object to be converted
    :param qtm_machine: (Optional) Quantinuum machine architecture to rebase against

    Returns:
        PHIR JSON as a str
    """
    logger.info("Starting phir conversion process for circuit %s", circuit)
    machine: Machine | None = None
    if qtm_machine:
        logger.info("Rebasing to machine %s", qtm_machine)
        circuit = rebase_to_qtm_machine(circuit, qtm_machine.value)
        machine = QTM_MACHINES_MAP.get(qtm_machine)
    else:
        machine = None

    logger.debug("Sharding input circuit...")
    sharder = Sharder(circuit)
    shards = sharder.shard()

    if machine:
        # Only print message if a machine object is passed
        # Otherwise, placement and routing are functionally skipped
        # The function is called, but the output is just filled with 0s
        logger.debug("Performing placement and routing...")
    placed = place_and_route(shards, machine)

    phir_json = genphir(placed, machine_ops=bool(machine))

    if logger.getEffectiveLevel() <= logging.INFO:
        print(PHIRModel.model_validate_json(phir_json))  # type: ignore[misc]
    return phir_json



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

import logging

import pytest

from pytket.phir.api import pytket_to_phir
from pytket.phir.qtm_machine import QtmMachine

from .sample_data import QasmFile, get_qasm_as_circuit

logger = logging.getLogger(__name__)


class TestApi:
    def test_pytket_to_phir_no_machine(self) -> None:
        """Test case when no machine is present."""
        circuit = get_qasm_as_circuit(QasmFile.baby)

        assert pytket_to_phir(circuit)

    @pytest.mark.parametrize("test_file", list(QasmFile))
    def test_pytket_to_phir_no_machine_all(self, test_file: QasmFile) -> None:
        """Test case when no machine is present."""
        circuit = get_qasm_as_circuit(test_file)

        match test_file:
            case QasmFile.big_gate:
                with pytest.raises(KeyError, match=r".*CnX.*"):
                    assert pytket_to_phir(circuit)
            case QasmFile.qv20_0:
                with pytest.raises(KeyError, match=r".*U3.*"):
                    assert pytket_to_phir(circuit)
            case _:
                assert pytket_to_phir(circuit)

    @pytest.mark.parametrize("test_file", list(QasmFile))
    def test_pytket_to_phir_h1_1_all(self, test_file: QasmFile) -> None:
        """Standard case."""
        circuit = get_qasm_as_circuit(test_file)

        assert pytket_to_phir(circuit, QtmMachine.H1_1)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

from argparse import ArgumentParser
from importlib.metadata import version

from pecos.engines.hybrid_engine import HybridEngine  # type:ignore [import-not-found]

from phir.model import PHIRModel
from pytket.qasm.qasm import (
    circuit_from_qasm,
    circuit_from_qasm_str,
    circuit_to_qasm_str,
)

from .api import pytket_to_phir
from .qtm_machine import QtmMachine
from .rebasing.rebaser import rebase_to_qtm_machine


def main() -> None:
    """pytket-phir compiler CLI."""
    parser = ArgumentParser(
        prog="phirc",
        description="Emulates QASM program execution via PECOS",
    )
    parser.add_argument(
        "qasm_files", nargs="+", default=None, help="One or more QASM files to emulate"
    )
    parser.add_argument(
        "-m",
        "--machine",
        choices=["H1-1", "H1-2"],
        default="H1-1",
        help="machine name, H1-1 by default",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f'{version("pytket-phir")}',
    )
    args = parser.parse_args()

    for file in args.qasm_files:
        print(f"Processing {file}")  # noqa: T201
        c = circuit_from_qasm(file)
        rc = rebase_to_qtm_machine(c, args.machine)
        qasm = circuit_to_qasm_str(rc, header="hqslib1")
        circ = circuit_from_qasm_str(qasm)

        match args.machine:
            case "H1-1":
                machine = QtmMachine.H1_1
            case "H1-2":
                machine = QtmMachine.H1_2
        phir = pytket_to_phir(circ, machine)
        PHIRModel.model_validate_json(phir)

        HybridEngine(qsim="state-vector").run(program=phir, shots=10)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from enum import Enum, auto
from pathlib import Path

from pytket.circuit import Circuit
from pytket.qasm.qasm import circuit_from_qasm


class QasmFile(Enum):
    baby = auto()
    simple = auto()
    eztest = auto()
    baby_with_rollup = auto()
    big_gate = auto()
    simple_cond = auto()
    n10_test = auto()
    classical_hazards = auto()
    cond_1 = auto()
    barrier_complex = auto()
    cond_classical = auto()
    bv_n10 = auto()
    oned_brickwork_circuit_n20 = auto()
    qv20_0 = auto()


def get_qasm_as_circuit(qasm_file: QasmFile) -> Circuit:
    """Utility function to convert a QASM file to Circuit.

    Args:
        qasm_file: enum for a QASM file

    Returns:
        Corresponding tket circuit
    """
    this_dir = Path(Path(__file__).resolve()).parent
    return circuit_from_qasm(f"{this_dir}/data/qasm/{qasm_file.name}.qasm")
